--
-- Multi variable monomials
-- A monomial is a term of variables x0, ..., xn each raised to a non-negative
-- integer.
-- Sub-library of Polynomial.hs
--

module Monomial (
	Variable,
	Monomial, unMono, mono, getExp, varm, applm, -- data type
	monoass1, monoass2, monoass3,                -- assertions
	divides, monoLift, multm, divm, lcmm, gcdm   -- operations
) where

import Control.Monad
import Data.Function
import Data.Monoid
import Data.List


------------------------------------
-- Data type
------------------------------------

type Variable = String

data Monomial = Mono { unMono :: [(Variable, Integer)] } deriving Eq
-- A monomial is a list of pairs of variables and their exponent
-- e.g. [("x", 1), ("y", 5)] corresponds to x*y^5

-- | A safe constructor for monomials (guarantees assertions)
mono :: [(Variable, Integer)] -> Monomial
mono []     = Mono []
mono (x@(v,i):xs)
	| i <  0 = error "Monomial.mono: negative exponent."
	| i == 0 = mono xs
	| i >  0 && v `elem` map fst xs' =
		Mono $ map (\(v',i') -> if v == v' then (v,i + i') else (v,i')) xs'
	| otherwise = Mono $ sortBy (compare `on` fst) $ (v,i):xs'
	where xs' = unMono $ mono xs

-- | Single variable constructor
varm :: Variable -> Monomial
varm v = Mono [(v, 1)]

-- | Get a variable's exponent in the monomial
getExp :: Monomial -> Variable -> Integer
getExp m = maybe 0 id . flip lookup (unMono m)

-- | Convert a monomial into a function
applm :: Num t => (Variable -> t) -> Monomial -> t
applm f = foldl (\x (v, i) -> f v ^ i * x) 1 . unMono

-- | Assertions for Monomials
monoass1, monoass2, monoass3 :: Monomial -> Bool
monoass1 = all ((> 0) . snd) . unMono             -- all exponents are positive
monoass2 = (\l -> l == nub l) . map fst . unMono  -- no duplicate entries
monoass3 = (\l -> l == sort l) . map fst . unMono -- sorted by variable names


------------------------------------
-- Class instances
------------------------------------

instance Show Monomial where
	show = intercalate "*"
		. (\l -> if null l then ["1"] else l)
		. map (\(v,i) -> v ++ if i > 1 then "^" ++ show i else "")
		. unMono

instance Monoid Monomial where
	mempty = Mono []
	mappend = multm


------------------------------------
-- Operations
------------------------------------

-- | Division test for Monomials
divides :: Monomial -> Monomial -> Bool
divides = (all snd .) . monoZip' (<=) `on` unMono

-- | Lift a function inside a monomial constructor
monoLift :: ([(String, Integer)] -> [(String, Integer)]) -> Monomial -> Monomial
monoLift f = mono . f . unMono

-- | zipWith functional lifted inside the monomial constructor
monoZip :: (Integer -> Integer -> Integer) -> Monomial -> Monomial -> Monomial
monoZip f = (mono .) . monoZip' f `on` unMono

monoZip' :: (Integer -> Integer -> a) -> [(String, Integer)] -> [(String, Integer)] -> [(String, a)]
monoZip' f []  []  = []
monoZip' f xs1 []  = map (\(v,i) -> (v, f i 0)) xs1
monoZip' f []  xs2 = map (\(v,i) -> (v, f 0 i)) xs2
monoZip' f (x1@(v1, i1):xs1) (x2@(v2, i2):xs2)
	| v1 <  v2 = (v1, f i1 0)  : monoZip' f xs1 (x2:xs2)
	| v1 == v2 = (v1, f i1 i2) : monoZip' f xs1 xs2
	| v1 >  v2 = (v2, f 0  i2) : monoZip' f (x1:xs1) xs2

-- | Multiplication for monomials
multm :: Monomial -> Monomial -> Monomial
multm = monoZip (+)

-- | Division for Monomials
divm :: Monomial -> Monomial -> Monomial
divm = monoZip (-)

-- | Least common multiple for monomials
lcmm :: Monomial -> Monomial -> Monomial
lcmm = monoZip max

-- | Greatest common divisor for monomials
gcdm :: Monomial -> Monomial -> Monomial
gcdm = monoZip min

