module Synthesis (
  K, Template, primeVar,
  ic_init, ic_cons, sc_post, sc_testcases, sc_testcase,
  template, sc, eq2smt,
  dosynthesis
) where

import Data.List
import Data.Monoid
import Data.Ratio

import LexOrder
import Monomial
import Polynomial
import Ideal
import ToSMT

type K = Rational -- some field; here: rationals, better: algebraic numbers
type Template = RationalFunction K
type TestCase = (Variable -> K, Variable -> K, [Ideal Template])


-- | Prefix of the invariant's template variables
inv_prefix :: String
inv_prefix = "inv" -- "psi"

-- | Transform a variable into a primed variable
primeVar :: Variable -> Variable
primeVar = ('\'' :) -- the prime is prepended because primed variables
                    -- should be ordered before unprimed variables

-- | Remove the prime from the variable, if primed
fromPrimeVar :: Variable -> Variable
fromPrimeVar v = if head v == '\'' then tail v else v

-- | Generate a list of all monomials in a generic polynomial of total degree d
varList :: [Variable] -> Integer -> [[Variable]]
varList vars 0 = []
varList vars 1 = [ [v]  | v <- vars ]
varList vars d = [ v : vl | v <- vars, vl <- vls, v <= head vl ] ++ varList vars 1
                 where vls = varList vars (d - 1)

-- | Generic template of total degree d
template :: String -> Integer -> [Variable] -> Polynomial Template
template prefix d vars = (+ cnst (var (prefix ++ "_0") % 1))
                    $ sum
                    $ map (\vl -> product (map var vl)
                                  * cnst (var (prefix ++ "_" ++ concat vl) % 1))
                    $ varList vars d


-----------------------------------------
-- Synthesis helper functions

-- | Generic template for Omega in the synthesis condition
omega :: Integer -> Polynomial Template
omega i = 1
-- omega i = template ("omega" ++ show i) 0 []

-- | Remove parameters from a template that are zero according to the equations
rmzeros :: Polynomial Template -> Assert Template -> Polynomial Template
rmzeros psi = foldr (\v -> clift (\f -> subst v 0 (numerator f) % denominator f)) psi
            . map (head . getVars . numerator)
            . filter poly_simple
            . flatten
  where
    -- | Is this a linear polynomial in one variable with zero constant?
    poly_simple :: Template -> Bool
    poly_simple t = (length $ getVars $ numerator t) == 1 && denominator t == 1

-- | Extract a system of equations in accordance with the Zero Polynomial Theorem
extract :: [Polynomial Template] -> Assert Template
extract = And . map (Eq 0 . fst) . concatMap coeff


-----------------------------------------
-- Invariance condition

-- | Initial condition in the invariance condition
ic_init :: Ideal Template -> Polynomial Template -> Assert Template
ic_init init psi = extract [snd $ division psi $ basis init]

-- | Consecution condition in the invariance condition
ic_cons :: [Variable] -> Polynomial Template -> [Ideal Template]
           -> Polynomial Template -> Polynomial Template -> Assert Template
ic_cons vars q transitions s psi = extract
                 $ map (\(tau, i) -> r tau - phi i tau * psi)
                 $ zip transitions [1..]
  where
    r tau     = snd (division (q * s * psi') (basis tau))
    phi i tau = template ("phi" ++ show i)
                         (total_deg (r tau) - total_deg psi)
                         vars
    psi'      = renameVars primeVar psi

-- | The invariance condition is not trivial (= 0)
ic_non_trivial :: Polynomial Template -> Assert Template
ic_non_trivial = Not . And . map (Eq 0 . fst) . coeff


-----------------------------------------
-- Synthesis condition

-- | Post condition in the synthesis condition
sc_post :: Ideal Template -> Ideal Template -> Polynomial Template -> Assert Template
sc_post exit post psi = extract [snd $ division (p - omega i * psi) $ basis exit
                                | (p, i) <- zip (basis post) [1..]]

-- | Constraints generated by test cases
sc_testcases :: Ideal Template -> Ideal Template -> [TestCase] -> Assert Template
sc_testcases init exit testcases = And $ map (sc_testcase init exit) testcases

-- | Generate the equations for one test case
sc_testcase :: Ideal Template -> Ideal Template -> TestCase -> Assert Template
sc_testcase init exit (pre, post, translist) =
  And (extract [snd $ division p $ state pre | p <- basis init]
       : extract [snd $ division p $ state post | p <- basis exit]
       : (if length (basis exit) == 0 then []
         else [Not $ extract [snd $ division p [var v - cnst (val' v) | v <- vars]
                             | p <- basis exit]
              | val' <- tail $ scanr update ((% 1) . cnst . pre) translist])
       ++ [Eq (val v) (cnst (post v) % 1) | v <- vars,
                          val <- [foldr update ((% 1) . cnst . pre) translist]])
  where
    vars = concatMap (filter ((/= '\'') . head) . concatMap getVars . basis) (init:translist)
    state val = [ var v - cnst (cnst (val v) % 1) | v <- vars ]
    update :: Ideal Template -> (Variable -> Template) -> Variable -> Template
    update trans val = extractVal
                       $ foldr (.) id [map (subst v $ cnst $ val v) | v <- vars]
                       $ basis
                       $ trans
    extractVal :: [Polynomial Template] -> Variable -> Template
    extractVal []     _ = undefined
    extractVal (p:ps) v | primeVar v `occursIn` p && poly_simple2 p
                                    = -fst (coeff p !! 1) / fst (coeff p !! 0)
                        | otherwise = extractVal ps v
    poly_simple2 p = onevar p
                     && length (coeff p) == 2
                     && length (unMono $ snd $ head $ coeff p) == 1
                     && (snd $ head $ unMono $ snd $ head $ coeff p) == 1
--------- the code below works, but is way too slow!
{-    update trans = map (renameVars unprimedVar)
                   . projectExtra
                   . basis
                   . ideal
                   . map (renameVars extraVar)
                   . (++ basis trans)
    projectExtra  = filter (any ((== '!') . head) . getVars) -- project away polynomials with !-variables
    extraVar v    = if head v == '\'' then v else '!' : v -- even larger monomials
    unprimedVar v = if head v == '\'' then tail v else v
-}

-- | Synthesis condition
sc :: [Variable] -> Polynomial Template -> Ideal Template
             -> Ideal Template -> Ideal Template
             -> [Ideal Template] -> [TestCase] -> Polynomial Template
             -> [Assert Template]
sc vars q init exit post transitions testcases psi = let
  s      = if length (basis exit) == 1 then head (basis exit) else 1
  a_init = ic_init init psi
--  psi2   = rmzeros psi a_init
  a_post = sc_post exit post psi
--  psi3   = rmzeros psi2 a_post
  a_cons = ic_cons vars q transitions s psi
  a_tc   = sc_testcases init exit testcases
  in [ic_non_trivial psi, a_init, a_post, a_cons, a_tc]


-----------------------------------------
-- Output and main

-- | Remove top Ands
removeAnd :: Assert Template -> [Assert Template]
removeAnd (And l) = concatMap removeAnd l
removeAnd x       = [x]

-- | Print the equations as an smt2 script that can be passed to an SMT solver
eq2smt :: [Assert Template] -> String
eq2smt asserts = "(set-option :produce-models true)\n(set-logic QF_NRA)\n\n"
             ++ concat ["(declare-fun " ++ v ++ " () Real)\n" | v <- vars]
             ++ "\n"
             ++ (concatMap (assert . toSMT)
--                 $ filter (not . trivial)
                 $ concatMap removeAnd asserts)
             ++ "\n(check-sat)\n(get-model)"
  where vars     = nub $ concatMap getVarsR $ concatMap flatten asserts
        assert s = "(assert " ++ s ++ ")\n"

-- | Main synthesis method combining everything
dosynthesis :: Polynomial Template -> Ideal Template -> Ideal Template -> Ideal Template
               -> [Ideal Template] -> [TestCase] -> Integer -> [Variable] -> IO ()
dosynthesis q init exit post transitions testcases d vars = putStrLn
              $ eq2smt
              $ sc vars q init exit post transitions testcases
              $ template inv_prefix d vars
