module ToSMT (
	ToSMT, toSMT,
	Assert(And, Or, Not, Eq), flatten, trivial
) where

import Data.List(intercalate)
import Data.Monoid
import Data.Ratio

import Monomial
import Polynomial

-- | The class of types that can be written as an SMT formula
class ToSMT a where
	toSMT :: a -> String

instance ToSMT Integer where
	toSMT n | n < 0     = "(- " ++ show (abs n) ++ ")"
	        | otherwise = show n

instance (ToSMT a, Integral a) => ToSMT (Ratio a) where
	toSMT r
		| denominator r == 1 = toSMT $ numerator r
		| otherwise          = "(/ " ++ toSMT (numerator r) ++ " "
		                       ++ toSMT (denominator r) ++ ")"

instance ToSMT Monomial where
	toSMT m
		| m == mempty = "1"
		| length (unMono m) == 1 = head $ bla m
		| otherwise   = "(* " ++ intercalate " " (bla m) ++ ")" where
			bla = concatMap (\(v, i) -> take (fromIntegral i) $ repeat v) . unMono

instance (ToSMT t, Eq t, Num t) => ToSMT (Polynomial t) where
	toSMT p
		| null (coeff p) = "0"
		| novar p = toSMT (fst $ head $ coeff p)
		| length (coeff p) == 1 = term2smt $ head $ coeff p
		| otherwise = "(+ " ++ (intercalate " " $ map term2smt $ coeff p) ++ ")"
		where
			term2smt (a,m) | a == 1 = toSMT m
			               | otherwise = "(* " ++ toSMT a ++ " " ++ toSMT m ++ ")"

data Assert t = And [Assert t] | Or [Assert t] | Not (Assert t) | Eq t t

instance ToSMT t => ToSMT (Assert t) where
	toSMT (And l) = "(and " ++ intercalate " " (map toSMT l) ++ ")"
	toSMT (Or l) = "(or " ++ intercalate " " (map toSMT l) ++ ")"
	toSMT (Not l) = "(not " ++ toSMT l ++ ")"
	toSMT (Eq x y) = "(= " ++ toSMT x ++ " " ++ toSMT y ++ ")"

instance ToSMT t => Show (Assert t) where
	show a = toSMT a

-- | Extract all occuring terms
flatten :: Assert t -> [t]
flatten (And l)  = concatMap flatten l
flatten (Or l)   = concatMap flatten l
flatten (Not x)  = flatten x
flatten (Eq x y) = [x, y]

-- | Is this a trivially satisfied assertion?
trivial :: Eq t => Assert t -> Bool
trivial (And l) = all trivial l
trivial (Or l) = any trivial l
trivial (Not x) = False -- 'not' not supported
trivial (Eq x y) = x == y
